# TensorFlow with Distributed Deep Learning Tutorial

## Overview

This tutorial explains the necessary steps for enabling Distributed Deep Learning (DDL) from within the `mnist.py` example provided in the PowerAI distribution.
DDL is indirectly integrated into TensorFlow in the form of a custom operator.
The custom operator is provided as a shared library, which can be loaded and invoked from within a python training script.

## Enabling DDL in a Tensorflow program

The DDL TensorFlow operator makes it easy to run TensorFlow programs on a cluster. Enabling DDL in your tensorflow program is simple and it can be done using one of the following two approaches:

* Enable DDL using the default options for the current system configuration
* Enable DDL manually

For both approaches to work most efficiently, the training data should be split up among each DDL instance. The DDL library provides functions to help with this.

### Enable DDL Using Default Options

In this approach, ddlrun figures out the best options to run with and sets `DDL_OPTIONS` environment variable. When the DDL library is imported in the TensorFlow program, the DDL library will be initialized using the `DDL_OPTIONS` environment variable. User can not override the environment variables using this approach.

The only change that is _required_ to enable DDL in this manner is to import the DDL python library:

```python
import ddl
```

The `mnist.py` script was modified in this manner and to split the training data, as described below, to enable DDL. The modified script can be found in `$CONDA_PREFIX/ddl-tensorflow/examples/mnist/mnist-env.py`.

### Enable DDL Manually

In this approach the DDL options are manually passed to the DDL library. When this approach is used, the `grads_reduce` function must be manually called. This gives the user more control over the use of the DDL library.

The following changes are _required_ to enable DDL in this manner:

1. Import the DDL Library

    ```python
    import ddl
    ```

1. Explicitly Initialize DDL

    ```python
    ddl.init(FLAGS.ddl_options)
    ```

1. Call DDL's grads_reduce Function

    Replace the reduceAll function in the mnist script with DDL's `grads_reduce` function.

    ```python
    grads_and_vars = ddl.grads_reduce(grads_and_vars, average=True)
    objective = optimizer.apply_gradients(grads_and_vars)
    ```

The mnist.py script was modified in this manner and to split the training data, as described below, to enable DDL. The modified script can be found in `$CONDA_PREFIX/ddl-tensorflow/examples/mnist/mnist-init.py`.

### Split the Training Data Among DDL Instances

For both approaches to work most efficiently, the data should be split up between each DDL instance. This can be done using the DDL functions `size()` and `rank()`. The `size()` function will return the total number of DDL instances, while `rank()` will return a unique integer for the current DDL instance.

Below is the code that was added to the `mnist.py` script, in both of the above approaches, in order to split up the data.

```python
batch_x, batch_y = mnist.train.next_batch(batch_size*ddl.size())

#select one of partitions
batch_x = np.split(batch_x,ddl.size())[ddl.rank()]
batch_y = np.split(batch_y,ddl.size())[ddl.rank()]
```

## Running a DDL Enabled TensorFlow Script

### Install ddl-tensorflow conda package

Before running a DDL enabled TensorFlow script, make sure there is a conda environment activated with the `ddl-tensorflow` package installed.

The mnist examples should also be copied into a user directory before being run. This can done be by exectuing the following command:

```bash
$ ddl-tensorflow-install-samples <somedir>
```

### Run Script Using ddlrun

DDL enabled programs should be launched using the `ddlrun` tool. See `$CONDA_PREFIX/doc/ddl/README.md` and `ddlrun -h` for more information on `ddlrun`.

#### Run Using DDL_OPTIONS

The `DDL_OPTIONS` environment variable should be set with the intended arguments. See `$CONDA_PREFIX/doc/ddl/README.md` for a list of DDL arguments.

The following command will launch the `mnist-env.py` script to run on host1 and host2.

```bash
ddlrun -H host1,host2 python mnist-env.py
```

#### Run Without Using DDL_OPTIONS

The DDL options can be passed into the `mnist-init.py` script as an argument. See `$CONDA_PREFIX/doc/ddl/README.md` for a list of DDL arguments.

The following command will launch the `mnist-init.py` script to run on host1 and host2.

```bash
ddlrun --no_ddloptions -H host1,host2 python mnist-init.py --ddl_options="-mode b:4x2"
```

&copy; Copyright IBM Corporation 2018, 2019
